@@ -31,25 +31,44 @@
 // CINT needs <algorithm> to work properly with <vector>
 #include <algorithm>
 
+#if HAVE_CONFIG_H
+# include "pynac-config.h"
+#endif
 #include "flags.h"
 #include "ptr.h"
 #include "assertion.h"
 #include "registrar.h"
 
+#include <factory/factory.h>
+
+#ifdef PYNAC_HAVE_LIBGIAC
+namespace giac
+{
+        class gen;
+        template <class T> class tensor;
+        typedef class tensor<gen> polynome;
+}
+namespace GiNaC
+{
+struct ex_is_less;
+}
+#endif
 
 namespace GiNaC {
 
 class ex;
-class ex_is_less;
+struct ex_is_less;
 class symbol;
 class numeric;
 class relational;
 class archive_node;
 class print_context;
+class flint_series_t;
 
 typedef std::vector<ex> exvector;
 typedef std::set<ex, ex_is_less> exset;
 typedef std::map<ex, ex, ex_is_less> exmap;
+using ex_int_map = std::map<GiNaC::ex, int, GiNaC::ex_is_less>;
 
 // Define this to enable some statistical output for comparisons and hashing
 #undef GINAC_COMPARE_STATISTICS
@@ -104,11 +123,30 @@ protected:
 /** This class is the ABC (abstract base class) of GiNaC's class hierarchy. */
 class basic : public refcounted
 {
-	GINAC_DECLARE_REGISTERED_CLASS_NO_CTORS(basic, void)
+public:
+	typedef void inherited;
+        static const tinfo_static_t tinfo_static;
+private:
+	static registered_class_info reg_info;
+public:
+	static registered_class_info &get_class_info_static() { return reg_info; }
+	virtual const registered_class_info &get_class_info() const { return basic::get_class_info_static(); }
+	virtual GiNaC::registered_class_info &get_class_info() { return basic::get_class_info_static(); }
+	virtual const char *class_name() const { return basic::get_class_info_static().options.get_name(); }
+
+	basic(const archive_node &n, lst &sym_lst);
+	virtual void archive(archive_node &n) const;
+	static ex unarchive(const GiNaC::archive_node &n, GiNaC::lst &sym_lst);
+
+	class visitor {
+	public:
+		virtual void visit(const basic &) = 0;
+		virtual ~visitor() {}; \
+	};
 
 	friend class ex;
-	friend struct print_order;
-	friend struct print_order_pair;
+	friend class print_order;
+	friend class print_order_pair;
 	// default constructor, destructor, copy constructor and assignment operator
 protected:
 	basic() : tinfo_key(&tinfo_static), flags(0) {}
@@ -157,7 +195,7 @@ public:
 
 	// operand access
 	virtual size_t nops() const;
-	virtual ex op(size_t i) const;
+	virtual const ex op(size_t i) const;
 	virtual ex operator[](const ex & index) const;
 	virtual ex operator[](size_t i) const;
 	virtual ex & let_op(size_t i);
@@ -203,9 +241,10 @@ protected:
 	virtual ex derivative(const symbol & s) const;
 public:
 	virtual ex series(const relational & r, int order, unsigned options = 0) const;
+        virtual void useries(flint_series_t& fp, int order) const {}
 
 	// rational functions
-	virtual ex normal(exmap & repl, exmap & rev_lookup, int level = 0) const;
+	virtual ex normal(exmap & repl, exmap & rev_lookup, int level = 0, unsigned options = 0) const;
 	virtual ex to_rational(exmap & repl) const;
 	virtual ex to_polynomial(exmap & repl) const;
 
@@ -282,6 +321,9 @@ public:
 
 protected:
 	void ensure_if_modifiable() const;
+#ifdef PYNAC_HAVE_LIBGIAC
+        const giac::polynome to_polynome(ex_int_map& map, exvector& revmap);
+#endif
 
 	// member variables
 protected:
@@ -312,6 +354,31 @@ inline bool is_exactly_a(const basic & obj)
 	return obj.tinfo() == &T::tinfo_static;
 }
 
+/** Constructs a new (class basic or derived) B object on the heap.
+ *
+ *  This function picks the object's ctor based on the given argument types.
+ *
+ *  This helps the constructor of ex from basic (or a derived class B) because
+ *  then the constructor doesn't have to duplicate the object onto the heap.
+ *  See ex::construct_from_basic(const basic &) for more information.
+ */
+template<class B, typename... Args>
+inline B & dynallocate(Args &&... args)
+{
+	return const_cast<B &>(static_cast<const B &>((new B(std::forward<Args>(args)...))->setflag(status_flags::dynallocated)));
+}
+/** Constructs a new (class basic or derived) B object on the heap.
+ *
+ *  This function is needed for GiNaC classes which have public ctors from
+ *  initializer lists of expressions (which are not a type and not captured
+ *  by the variadic template version).
+ */
+template<class B>
+inline B & dynallocate(std::initializer_list<ex> il)
+{
+	return const_cast<B &>(static_cast<const B &>((new B(il))->setflag(status_flags::dynallocated)));
+}
+
 } // namespace GiNaC
 
 #endif // ndef __GINAC_BASIC_H__
